# -*- python -*-
# $RCSfile: gfxwindow.py,v $
# $Revision: 1.226 $
# $Author: langer $
# $Date: 2006/08/29 21:34:36 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

import gobject
import gtk
from types import *
import math, time, sys

from oof2.SWIG.common import coord
from oof2.SWIG.common import geometry
from oof2.SWIG.common import lock
from oof2.SWIG.common.IO.GUI import oofcanvas
#from oof2.common.IO.GUI import fakecanvas
from oof2.SWIG.common import threadstate
from oof2.SWIG.common.IO.GUI import rubberband
from oof2.common import debug
from oof2.common import mainthread
from oof2.common import primitives
from oof2.common import subthread
from oof2.common import toolbox
from oof2.common import utils
from oof2.common.IO import ghostgfxwindow
from oof2.common.IO import gfxmanager
from oof2.common.IO import layereditor
from oof2.common.IO import mainmenu
from oof2.common.IO import reporter
from oof2.common.IO.GUI import chooser
from oof2.common.IO.GUI import fileselector
from oof2.common.IO.GUI import gfxmenu
from oof2.common.IO.GUI import gtklogger
from oof2.common.IO.GUI import gtkutils
from oof2.common.IO.GUI import mousehandler
from oof2.common.IO.GUI import quit
from oof2.common.IO.GUI import subWindow
from oof2.common.IO.GUI import toolboxGUI
from oof2.common.guitop import top


# during_callback() is called (by CanvasOutput.show()) only in
# non-threaded mode, so we don't worry about the thread-safety of a
# global variable here.
_during_callback = 0
def during_callback():
    return _during_callback

_pi_over_2 = math.pi/2.0

class ContourMapData:
    # This class just aggregates data related to the ContourMap
    # display, to keep the code (relatively) tidy.
    def __init__(self):
        self.canvas = None
        self.rawdevice = None
        self.device = None
        self.mouse_down = None
        self.mark_value = None
        self.canvas_mainlayer = None
        self.canvas_ticklayer = None

class GfxWindow(subWindow.SubWindow, ghostgfxwindow.GhostGfxWindow):
    # TODO LATER: Set initial sizes in some more user-friendly way.
    initial_height = 400
    initial_width = 800
    def __init__(self, name, count, gfxmanager, clone=0):
        self.gfxlock = lock.Lock()
        mainthread.runBlock(self.preinitialize, (name, count,
                                                gfxmanager, clone))
        self.newCanvas()
        ghostgfxwindow.GhostGfxWindow.__init__(self, name, count,
                                               gfxmanager, clone=clone)
        mainthread.runBlock(self.postinitialize, (name, count,
                                                  gfxmanager, clone))

        # This whole initialization sequence is complicated.  The
        # reason is that the order of operations is important -- the
        # ghostgfxwindow makes function calls which need to be
        # redefined in the GUI before they occur -- and there is a
        # requirement that the main thread never be locked.
        # Furthermore, the ghostgfxwindow has to create the menu
        # before the SubWindow init gets called.  It could probably be
        # rationalized, but it's not urgent.


    def preinitialize(self, name, count, gfxmanager, clone):
        debug.mainthreadTest()
        self.gtk = None
        self.closed = None # State data used at window-close time.
        self.name = name
        self.oofcanvas = None
        self.realized = 0
        self.zoomed = 0
        self.settings = ghostgfxwindow.GfxSettings()
        self.mouseHandler = mousehandler.nullHandler # doesn't do anything
        self.rubberband = rubberband.NoRubberBand()

        self.contourmapdata = ContourMapData()

        # Build all the GTK objects for the interior of the box.  These
        # actually get added to the window itself after the SubWindow
        # __init__ call.  They need to be created first so the
        # GhostGfxWindow can operate on them, and then create the menus
        # which are handed off to the SubWindow.
        self.mainpane = gtk.VPaned()
        gtklogger.setWidgetName(self.mainpane, 'Pane0')

        # Panes dividing upper pane horizontally into 3 parts.
        # paned1's left half contains paned2.
        self.paned1 = gtk.HPaned()
        gtklogger.setWidgetName(self.paned1, "Pane1")
        self.mainpane.pack1(self.paned1, resize=True)
        gtklogger.connect_passive(self.paned1, 'size-allocate')

        # paned2 is in left half of paned1
        self.paned2 = gtk.HPaned()
        gtklogger.setWidgetName(self.paned2, "Pane2")
        self.paned1.pack1(self.paned2, resize=True)
        gtklogger.connect_passive(self.paned2, 'size-allocate')

        # The toolbox is in the left half of paned2 (ie the left frame of 3)
        toolboxframe = gtk.Frame()
        toolboxframe.set_shadow_type(gtk.SHADOW_IN)
        self.paned2.pack1(toolboxframe, resize=True)
        gtklogger.connect_passive(toolboxframe, 'size-allocate')

        # Box containing the toolbox label and the scroll window for
        # the toolbox itself.
        toolboxbox1 = gtk.VBox()
        toolboxframe.add(toolboxbox1)
        hbox = gtk.HBox()
        toolboxbox1.pack_start(hbox, expand=0, fill=0, padding=2)
        hbox.pack_start(gtk.Label("Toolbox:"), expand=0, fill=0, padding=3)
        
        self.toolboxchooser = chooser.ChooserWidget([],
                                                    callback=self.switchToolbox,
                                                    name="TBChooser")
        hbox.pack_start(self.toolboxchooser.gtk, expand=1, fill=1, padding=3)

        # Scroll window for the toolbox itself.
        toolboxbox2 = gtk.ScrolledWindow()
        gtklogger.logScrollBars(toolboxbox2, 'TBScroll')
        
        toolboxbox2.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        toolboxbox1.pack_start(toolboxbox2, expand=1, fill=1)

        # Actually, the tool box goes inside yet another box, so that
        # we have a gtk.VBox that we can refer to later.
        self.toolboxbody = gtk.VBox()
        toolboxbox2.add_with_viewport(self.toolboxbody)

        self.toolboxGUIs = []           # GUI wrappers for toolboxes.
        self.current_toolbox = None

        # The canvas is in the right half of paned2 (ie the middle
        # pane of 3).  We *don't* use a ScrolledWindow for it, because
        # we need direct access to the Scrollbars.  Instead, we make
        # the Scrollbars ourselves and put them in a Table with the
        # canvas.
        self.canvasTable = gtk.Table(rows=2, columns=2)
        gtklogger.setWidgetName(self.canvasTable, "Canvas")
        self.canvasTable.set_col_spacings(0)
        self.canvasTable.set_row_spacings(0)
        frame = gtk.Frame()
        frame.set_shadow_type(gtk.SHADOW_IN)
        frame.add(self.canvasTable)
        self.paned2.pack2(frame, resize=True)
        self.hScrollbar = gtk.HScrollbar()
        self.vScrollbar = gtk.VScrollbar()
        gtklogger.setWidgetName(self.hScrollbar, "hscroll")
        gtklogger.setWidgetName(self.vScrollbar, "vscroll")
        # Catch button release events on the Scrollbars, so that their
        # changes can be logged.  *Don't* catch the "changed" signals
        # from their Adjustments, because they occur too often.
        gtklogger.connect(self.hScrollbar, "button-release-event",
                          self.scrlReleaseCB, 'h')
        gtklogger.connect(self.vScrollbar, "button-release-event",
                          self.scrlReleaseCB, 'v')
        self.canvasTable.attach(self.hScrollbar, 0,1, 1,2,
                                xoptions=gtk.EXPAND|gtk.FILL, yoptions=gtk.FILL)
        self.canvasTable.attach(self.vScrollbar, 1,2, 0,1,
                                xoptions=gtk.FILL, yoptions=gtk.EXPAND|gtk.FILL)
        self.canvasFrame = gtk.Frame()
        self.canvasFrame.set_shadow_type(gtk.SHADOW_NONE)
        self.canvasTable.attach(self.canvasFrame, 0,1, 0,1)

        # self.newCanvas()

        # the contourmap is in the right half of paned1 (the right pane of 3)
        contourmapframe = gtk.Frame()
        contourmapframe.set_shadow_type(gtk.SHADOW_IN)
        self.paned1.pack2(contourmapframe, resize=False)

        contourmapbox = gtk.VBox()
        gtklogger.setWidgetName(contourmapbox, "ContourMap")
        contourmapframe.add(contourmapbox)

        self.contourmap_max = gtk.Label("max")
        gtklogger.setWidgetName(self.contourmap_max, "MapMax")
        self.new_contourmap_canvas()    # Sets self.contourmapdata.canvas.
        self.contourmap_min = gtk.Label("min")
        gtklogger.setWidgetName(self.contourmap_min, "MapMin")
        
        contourmaplevelbox = gtk.HBox()
        self.contourlevel_min = gtk.Label("   min   ")
        contourmaplevelbox.pack_start(self.contourlevel_min, True, True)
        contourmaplevelbox.pack_start(gtk.VSeparator(), False, False)
        contourmaplevelbox.pack_start(gtk.VSeparator(), False, False)
        self.contourlevel_max = gtk.Label("   max   ")
        contourmaplevelbox.pack_end(self.contourlevel_max, True, True)
        
        contourmapclearbutton = gtk.Button("Clear Mark")
        gtklogger.setWidgetName(contourmapclearbutton, "Clear")
        gtklogger.connect(contourmapclearbutton, "clicked",
                         self.contourmap_clear_marker)
        
        contourmapbox.pack_start(self.contourmap_max, False, False)
        contourmapbox.pack_start(gtk.HSeparator(), False, False)
        contourmapbox.pack_start(self.contourmapdata.canvas.widget(), True, True)
        contourmapbox.pack_start(gtk.HSeparator(), False, False)
        contourmapbox.pack_start(self.contourmap_min, False, False)

        contourmapbox.pack_start(gtk.HSeparator(), False, False)
        contourmapbox.pack_start(gtk.HSeparator(), False, False)
        contourmapbox.pack_start(contourmaplevelbox, False, False)
        contourmapbox.pack_start(gtk.HSeparator(), False, False)
        contourmapbox.pack_end(contourmapclearbutton, False, False)

        contourmapframe.show_all()


        # HACK.  Set the position of the toolbox/canvas divider.  This
        # prevents the toolbox pane from coming up minimized.
        self.paned2.set_position(250)

        # Bottom part of main pane is a list of layers.  The actual
        # DisplayLayer objects are stored in self.display.

        layerFrame = gtk.Frame(label='Layers')
        
        self.mainpane.pack2(layerFrame, resize=False)
        self.layerScroll = gtk.ScrolledWindow()
        gtklogger.logScrollBars(self.layerScroll, "LayerScroll")
        self.layerScroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        layerFrame.add(self.layerScroll)

        self.layerList = gtk.ListStore(gobject.TYPE_PYOBJECT)
        self.layerListView = gtk.TreeView(self.layerList)
        gtklogger.setWidgetName(self.layerListView, "LayerList")
        self.layerListView.set_row_separator_func(self.layerRowSepFunc)
        self.layerListView.set_reorderable(True)
        self.layerListView.set_fixed_height_mode(False)
        self.layerScroll.add(self.layerListView)

        gtklogger.adoptGObject(self.layerList, self.layerListView,
                              access_method=self.layerListView.get_model)

        # The row-deleted and row-inserted signals are used to detect
        # when the user has reordered rows manually.  When the program
        # does anything that might cause these signals to be emitted,
        # it must first call suppressRowOpSignals.
        self.rowOpSignals = [
            gtklogger.connect(self.layerList, "row-deleted",
                             self.listRowDeletedCB),
            gtklogger.connect(self.layerList, "row-inserted",
                             self.listRowInsertedCB)
            ]
        self.destination_path = None

        showcell = gtk.CellRendererToggle()
        showcol = gtk.TreeViewColumn("Show")
        showcol.pack_start(showcell, expand=False)
        showcol.set_cell_data_func(showcell, self.renderShowCell)
        self.layerListView.append_column(showcol)
        gtklogger.adoptGObject(showcell, self.layerListView,
                               access_function='findCellRenderer',
                               access_kwargs={'col':0, 'rend':0})
        gtklogger.connect(showcell, 'toggled', self.showcellCB)        

        cmapcell = gtk.CellRendererToggle()
        cmapcell.set_radio(True)
        cmapcol = gtk.TreeViewColumn("Map")
        cmapcol.pack_start(cmapcell, expand=False)
        cmapcol.set_cell_data_func(cmapcell, self.renderCMapCell)
        self.layerListView.append_column(cmapcol)
        gtklogger.adoptGObject(cmapcell, self.layerListView,
                               access_function='findCellRenderer',
                               access_kwargs={'col':1, 'rend':0})
        gtklogger.connect(cmapcell, 'toggled', self.cmapcellCB)        

        layercell = gtk.CellRendererText()
        layercol = gtk.TreeViewColumn("What")
        layercol.set_resizable(True)
        layercol.pack_start(layercell, expand=True)
        layercol.set_cell_data_func(layercell, self.renderLayerCell)
        self.layerListView.append_column(layercol)

        methodcell = gtk.CellRendererText()
        methodcol = gtk.TreeViewColumn("How")
        methodcol.set_resizable(True)
        methodcol.pack_start(methodcell, expand=True)
        methodcol.set_cell_data_func(methodcell, self.renderMethodCell)
        self.layerListView.append_column(methodcol)

        gtklogger.adoptGObject(self.layerListView.get_selection(),
                              self.layerListView,
                              access_method=self.layerListView.get_selection)
        self.selsignal = gtklogger.connect(self.layerListView.get_selection(), 
                                          'changed', self.selectionChangedCB)
        gtklogger.connect(self.layerListView, 'row-activated',
                         self.layerDoubleClickCB)
        



    def postinitialize(self, name, count, gfxmanager, clone):
        debug.mainthreadTest()
        # Add gui callbacks to the non-gui menu created by the GhostGfxWindow.
        filemenu = self.menu.File
        filemenu.Save_Image.add_gui_callback(self.saveImage_gui)
        filemenu.Save_Contourmap.add_gui_callback(self.saveContourmap_gui)
        filemenu.Quit.add_gui_callback(quit.quit)
        layermenu = self.menu.Layer
        layermenu.New.add_gui_callback(self.newLayer_gui)
        layermenu.Edit.add_gui_callback(self.editLayer_gui)
        layermenu.Delete.add_gui_callback(self.deleteLayer_gui)
        layermenu.Hide.add_gui_callback(self.hideLayer_gui)
        layermenu.Show.add_gui_callback(self.showLayer_gui)
        layermenu.Hide_Contour_Map.add_gui_callback(
            self.hideLayerContourmap_gui)
        layermenu.Show_Contour_Map.add_gui_callback(
            self.showLayerContourmap_gui)
        layermenu.Raise.One_Level.add_gui_callback(self.raiseLayer_gui)
        layermenu.Raise.To_Top.add_gui_callback(self.raiseToTop_gui)
        layermenu.Lower.One_Level.add_gui_callback(self.lowerLayer_gui)
        layermenu.Lower.To_Bottom.add_gui_callback(self.lowerToBottom_gui)
        settingmenu = self.menu.Settings
        toolboxmenu = self.menu.Toolbox

        # Construct gui's for toolboxes.  This must be done after the
        # base class is constructed so that all non-gui toolboxes are
        # created before any of their gui versions.  Some gui
        # toolboxes need to know about more than one non-gui toolbox.
        map(self.makeToolboxGUI, self.toolboxes)
        if self.toolboxGUIs:
            self.selectToolbox(self.toolboxGUIs[0].name())
            self.toolboxchooser.set_state(self.toolboxGUIs[0].name())

        # raise_window routine is in SubWindow class.
        getattr(mainmenu.OOF.Windows.Graphics, name).add_gui_callback(
            self.raise_window)

        # SubWindow initializer makes the menu bar, and sets up the
        # .gtk and .mainbox members.  ".gtk" is the window itself,
        # and .mainbox is a gtk.VBox that holds the menu bar.
        windowname = utils.underscore2space("OOF2 " + name)
        subWindow.SubWindow.__init__(
            self, windowname, menu=self.menu)



        self.gtk.connect('destroy', self.destroyCB)
        self.gtk.connect_after('realize', self.realizeCB)
        self.gtk.set_default_size(GfxWindow.initial_width,
                                  GfxWindow.initial_height)

        self.mainbox.set_spacing(3)

        self.mainbox.pack_start(self.mainpane, fill=1, expand=1)

        self.gtk.show_all()

        self.updateToolboxChooser()

    def __repr__(self):
        return 'GfxWindow("%s")' % self.name

    ################################################

    def newCanvas(self):
        # Recreate the canvas object.
        # It's important to acquire and release the lock in the
        # subthread, before calling mainthread.runBlock, to avoid
        # deadlocks.
        self.acquireGfxLock()
        try:
            mainthread.runBlock(self.newCanvas_thread)
        finally:
            self.releaseGfxLock()
    def newCanvas_thread(self):
        debug.mainthreadTest()
        ppu = None
        if self.oofcanvas:
            ppu = self.oofcanvas.get_pixels_per_unit()
            scrollregion = self.oofcanvas.get_scrollregion()
            offsets = self.oofcanvas.get_scroll_offsets()
            self.oofcanvas.widget().destroy()

        self.oofcanvas = oofcanvas.OOFCanvas(self.settings.antialias)
##        self.oofcanvas = fakecanvas.FakeCanvas(self.settings.antialias)
        self.canvasFrame.add(self.oofcanvas.widget())
        self.hScrollbar.set_adjustment(self.oofcanvas.get_hadjustment())
        self.vScrollbar.set_adjustment(self.oofcanvas.get_vadjustment())
        # Changes to the adjustments need to go into the gui log.
        gtklogger.adoptGObject(self.hScrollbar.get_adjustment(),
                               self.hScrollbar,
                               access_method=self.hScrollbar.get_adjustment)
        gtklogger.adoptGObject(self.vScrollbar.get_adjustment(),
                               self.vScrollbar,
                               access_method=self.vScrollbar.get_adjustment)
        gtklogger.connect_passive(self.hScrollbar.get_adjustment(),
                                  'value-changed')
        gtklogger.connect_passive(self.vScrollbar.get_adjustment(),
                                  'value-changed')

        # GUI logging stuff.
        canvasroot = self.oofcanvas.rootitem()
        init_canvas_logging(canvasroot) # one-time initialization
        # Although canvasroot is a gtk widget, it's not a pygtk
        # widget, which makes life difficult.  So here it's adopted by
        # a pygtk widget instead, and uses various hacks to get itself
        # logged.  The actual widget doing the adopting is completely
        # arbitrary since it will be passed as the first argument of
        # findCanvasRoot, which discards it.  It does have to be a
        # loggable pygtk widget, though, because
        # AdopteeLogger.location() doesn't know that it will be
        # discarded.  I did say that this is a hack, didn't I?
        gtklogger.adoptGObject(canvasroot, self.canvasTable,
                               access_function='findCanvasRoot',
                               access_kwargs={"windowname":self.name})
        gtklogger.connect_passive(canvasroot, "event")

        if ppu is not None:
            self.oofcanvas.set_pixels_per_unit(ppu)
            self.oofcanvas.set_scrollregion(scrollregion)
            self.oofcanvas.set_scroll_offsets(offsets)

        # delayed import to avoid import loops
        from oof2.common.IO.GUI import canvasoutput
        from oof2.common.IO import outputdevice

        # Use the buffered device.
        rawdevice = canvasoutput.CanvasOutput(self.oofcanvas)
        self.device = outputdevice.BufferedOutputDevice(rawdevice)

        # On a new canvas, these should both be set at once,
        # since they both need to call underlay.
        self.oofcanvas.set_underlay_params(
            self.settings.bgcolor, self.settings.margin)
        self.oofcanvas.set_mouse_callback(self.mouseCB)
        self.oofcanvas.set_rubberband(self.rubberband)
        self.oofcanvas.show()

        self.fix_step_increment()

    def get_oofcanvas(self):
        return self.oofcanvas

    def newLayerMembers(self):
        self.fillLayerList()
        ghostgfxwindow.GhostGfxWindow.newLayerMembers(self)

    ## Only call fillLayerList if the whole list needs to be rebuilt.
    def fillLayerList(self):
        mainthread.runBlock(self.fillLayerList_thread)
    def fillLayerList_thread(self):
        self.suppressRowOpSignals()
        self.suppressSelectionSignals()
        try:
            self.layerList.clear()
            for layer in self.display:
                self.layerList.prepend([layer])
                if layer is self.selectedLayer:
                    self.layerListView.get_selection().select_path('0')
        finally:
            self.allowRowOpSignals()
            self.allowSelectionSignals()
            
    #############

    # Callbacks for the TreeView for the Layer List.
    
    # TreeView callback for setting the state of the "Show" button 
    def renderShowCell(self, column, cell_renderer, model, iter):
        layer = model[iter][0]
        cell_renderer.set_active(not layer.hidden)
##        cell_renderer.set_property('active', not layer.hidden)        

    def showcellCB(self, cell_renderer, path):
        layer = self.layerList[path][0]
        if layer.hidden:
            self.menu.Layer.Show(n=self.layerID(layer))
        else:
            self.menu.Layer.Hide(n=self.layerID(layer))

    def renderCMapCell(self, column, cell_renderer, model, iter):
        layer = model[iter][0]
        if layer.contour_capable(self):
            cell_renderer.set_property('activatable', True)
            cell_renderer.set_active(layer is self.current_contourmap_method)
        else:
            cell_renderer.set_property('activatable', False)
            cell_renderer.set_property('active', False)

    def cmapcellCB(self, cell_renderer, path):
        layer = self.layerList[path][0]
        if layer is self.current_contourmap_method:
            self.menu.Layer.Hide_Contour_Map(n=self.layerID(layer))
        else:
            self.menu.Layer.Show_Contour_Map(n=self.layerID(layer))

    def renderLayerCell(self, column, cell_renderer, model, iter):
        layer = model[iter][0]
        who = layer.who()
        if who is not None:
            txt = "%s(%s)" % (who.getClassName(), who.name())
        else:
            # who is None.  This probably can't happen, but there was
            # comment here wondering if it could, and it's harmless to
            # check for it.
            txt = "???"
        cell_renderer.set_property('text', txt)

    def renderMethodCell(self, column, cell_renderer, model, iter):
        layer = model[iter][0]
        if self.settings.longlayernames:
            cell_renderer.set_property('text', `layer`)
        else:
            cell_renderer.set_property('text', layer.short_name())
                                   
    def selectionChangedCB(self, selection): # self.layerListView callback
        model, iter = selection.get_selected()
        self.suppressSelectionSignals()
        try:
            if iter is not None:
                layer = model[iter][0]
                self.menu.Layer.Select(n=self.layerID(layer))
            else:
                self.menu.Layer.Deselect(n=self.layerID(self.selectedLayer))
        finally:
            self.allowSelectionSignals()

    def layerDoubleClickCB(self, treeview, path, col):
        if self.selectedLayer is not None:
            layereditor.menu.LayerSet.Edit(
                window=self.name, layer_number=self.layerID(self.selectedLayer))

    # TreeView callback that determines if a row is displayed as a
    # separator.
    def layerRowSepFunc(self, model, iter):
        layer = model[iter][0]
        return not (layer.listed or self.settings.listall)

    # Callbacks for the layerList's "row-inserted" and "row-deleted"
    # signals, sent when the user reorders layers by dragging them in
    # the layer list.  "row-inserted" is sent first.
    
    def listRowInsertedCB(self, model, path, iter):
        # "row-inserted" is sent before the row is built, so we can't
        # get the actual layer here, just where it's going to be.
        self.destination_path = path
        self.suppressSelectionSignals()
    def listRowDeletedCB(self, model, path):
        self.allowSelectionSignals()
        if self.destination_path is not None:
            source_row = path[0]
            dest_row = self.destination_path[0]
            if source_row > dest_row:
                # The layer has been raised.  Remember that indices in the
                # gtk ListStore run in the opposite direction from indices
                # in the Display's layer list.
                layer = model[self.destination_path][0]
                # How far has the row moved? The -1 is because the path
                # was computed *before* the row was deleted.
                delta = source_row - dest_row - 1
                if delta > 0:
                    self.menu.Layer.Raise.By(n=self.layerID(layer),
                                             howfar=delta)
            else:                           # source_row < dest_row
                # The layer has been lowered.
                # After the source row is deleted, the destination row
                # number decreases by 1.
                layer = model[(dest_row-1,)][0]
                delta = dest_row - source_row - 1
                if delta > 0:
                    self.menu.Layer.Lower.By(n=self.layerID(layer),
                                             howfar=delta)
            self.destination_path = None

    # suppressRowOpSignals and allowRowOpSignals are used to make sure
    # that the row rearrangement callbacks aren't invoked when rows
    # are manipulated from the program, instead of by the user.
    def suppressRowOpSignals(self):
        for signal in self.rowOpSignals:
            signal.block()
    def allowRowOpSignals(self):
        for signal in self.rowOpSignals:
            signal.unblock()

    def suppressSelectionSignals(self):
        self.selsignal.block()
    def allowSelectionSignals(self):
        self.selsignal.unblock()
        
    # Call layerListRowChanged to notify the TreeView that the data it
    # displays has changed.
    def layerListRowChanged(self, n):
        rowno = len(self.layerList) - 1 - n
        self.layerList.row_changed(rowno, self.layerList.get_iter(rowno))

    ##############
    
    def replaceLayer(self, count, oldlayer, newlayer):
        self.acquireGfxLock()
        try:
            ghostgfxwindow.GhostGfxWindow.replaceLayer(self, count,
                                                       oldlayer, newlayer)
            for row in self.layerList:
                if row[0] is oldlayer:
                    row[0] = newlayer
                    break
        finally:
            self.releaseGfxLock()

    ###########################################
    # Color bar stuff.

    # Create object and assign to self.contourmap_canvas.
    def new_contourmap_canvas(self):
        mainthread.runBlock(self.new_contourmap_canvas_thread)
    def new_contourmap_canvas_thread(self):
        if self.contourmapdata.canvas:
            self.contourmapdata.canvas.destroy()

        # Argument is antialias parameter.
        self.contourmapdata.canvas = oofcanvas.OOFCanvas(0)

        # Duplicate imports, again to avoid an import loop.
        from oof2.common.IO.GUI import canvasoutput
        from oof2.common.IO import outputdevice
        
        self.contourmapdata.rawdevice=canvasoutput.CanvasOutput(
            self.contourmapdata.canvas)
        self.contourmapdata.device=outputdevice.BufferedOutputDevice(
            self.contourmapdata.rawdevice)
        
        self.contourmapdata.canvas.set_configure_callback(
            self.contourmap_config)
        self.contourmapdata.canvas.set_mouse_callback(self.contourmap_mouse)
        # Params are background color and margin.
        self.contourmapdata.canvas.set_underlay_params(
            self.settings.bgcolor, 0.0)

        # Create two layers, one for the "main" drawing, and
        # one for the ticks.
        self.contourmapdata.canvas_mainlayer = \
                               self.contourmapdata.device.begin_layer()
        self.contourmapdata.canvas_ticklayer = \
                               self.contourmapdata.device.begin_layer()
        self.contourmapdata.canvas_ticklayer.raise_to_top()
    
##    # Function called after layers have been arranged.  Does not
##    # draw the actual contourmap, just records the layers.
##    def contourmap_newlayers(self):
##        ghostgfxwindow.GhostGfxWindow.contourmap_newlayers(self)
##        for ell in self.display:
##            if ell==self.current_contourmap_method:
##                mainthread.runBlock(
##                    self.layerwidgets[ell].setContourmapButton,(1,))
##            else:
##                mainthread.runBlock(
##                    self.layerwidgets[ell].setContourmapButton,(0,))

    # def set_contourmap_layer(self, method):
    #     ghostgfxwindow.GhostGfxWindow.set_contourmap_layer(self, method)
        
            
    # Draw the contourmap onto the canvas, and update the numbers on the
    # display.  Called after all the layers have been drawn in the
    # main pane.  Also called from the menu callback which resets the
    # aspect ratio.

    # TODO LATER: There is some clunkiness in this routine, associated
    # particularly with the "discrete" displays, most particularly
    # with the CenterFill display.  The problem is that, in the
    # center-fill display and the skeleton quality display, the colors
    # drawn on the main canvas are broad swaths painted with the color
    # corresponding to a particular level, which behaves as a "contour
    # level" -- when there are few of these, this looks dumb, some
    # colors which may occur over a wide area of the main canvas will
    # be invisible on the contourmap.  This problem occurs for regular
    # contours also, but isn't as noticeable, partially because you
    # have more of them, and partially because the maximum of a real
    # contour plot occupies a small fraction of the main-canvas area.
    
    # A possible answer is to make the colormap be drawn continuously,
    # no matter how many levels their actually are.  Alternatively,
    # one could figure out how to do something sensible with the fact
    # that contour lines in general occur at particular levels, but if
    # you want to see something on the map, it has to be drawn over an
    # interval.

    # Called on a subthread now, but that's OK, device is buffered.
    def show_contourmap_info(self):
        if not self.gtk:
            return
        self.contourmapdata.canvas_mainlayer.clear()
        self.contourmapdata.canvas_mainlayer.make_current()
        
        c_min = None
        c_max = None

        # Copy self.current_contourmap_method to a local variable to
        # prevent interference from other threads.
        current_contourmethod = self.current_contourmap_method

        if current_contourmethod:
            current_contourmethod.draw_contourmap(
                self, self.contourmapdata.device)
            (c_min, c_max, lvls) = \
                    current_contourmethod.get_contourmap_info()


            # Flush the buffered device, so geometry data is valid.
            self.contourmapdata.device.flush_wait()
            
            # Zoom and scroll the canvas so that the drawn contour map 
            # exactly fills it vertically.
            ph = mainthread.runBlock(
                self.contourmapdata.canvas.get_height_in_pixels)

            # c_max can equal c_min....
            if c_max!=c_min:
                ppu = 1.0*ph/abs(c_max-c_min)
            else:
                ppu = ph  # Arbitrary, one unit for all the pixels.

            # Set the scrollregion and pixels-per-unit in the right
            # order.  If the new ppu is much larger than then old ppu,
            # the memory required by the canvas will explode unless
            # the new scrollregion set first.  It's also necessary to
            # call canvas.underlay() at the right time, so that a
            # large underlayer isn't drawn with a high ppu.
            oldppu = mainthread.runBlock(
                self.contourmapdata.canvas.get_pixels_per_unit)
            if ppu > oldppu:            # new map is smaller in physical units
                mainthread.runBlock(self.contourmapdata.canvas.underlay)
                # Recompute bounds *after* recomputing the underlayer,
                # because the old underlayer is too big, and is
                # included in the bounds computation.
                bounds = mainthread.runBlock(
                    self.contourmapdata.canvas.get_bounds)
                mainthread.runBlock(
                    self.contourmapdata.canvas.set_scrollregion, (bounds,))
                # Now that the canvas is only drawing small things,
                # it's ok to set ppu to something big.
                mainthread.runBlock(
                    self.contourmapdata.canvas.set_pixels_per_unit, (ppu,))
            else:           # ppu < oldppu, new map is larger in phyical units
                mainthread.runBlock(
                    self.contourmapdata.canvas.set_pixels_per_unit, (ppu,))
                bounds = mainthread.runBlock(
                    self.contourmapdata.canvas.get_bounds)
                mainthread.runBlock(
                    self.contourmapdata.canvas.set_scrollregion, (bounds,))
                mainthread.runBlock(
                    self.contourmapdata.canvas.underlay)

        self.contourmapdata.device.end_layer()
        self.contourmapdata.device.show()
        
        if c_min is None:
            mainthread.runBlock(
                self.contourmap_min.set_text, ('min',))
        else:
            mainthread.runBlock(
                self.contourmap_min.set_text,
                ( ("%.3g" % c_min).rstrip(), ) )
            
        if c_max is None:
            mainthread.runBlock(
                self.contourmap_max.set_text, ('max',) )
        else:
            mainthread.runBlock(
                self.contourmap_max.set_text,
                ( ("%.3g" % c_max).rstrip(), ) )

        self.show_contourmap_ticks(self.contourmapdata.mark_value)
        gtklogger.checkpoint("contourmap info updated for " + self.name)


    # Draw the marks on it.  Argument is the new value for the ticks.
    # A tick-layer redraw can be forced by setting
    # self.contourmapdata.mark_value to None and then calling this with
    # the new value.
    def show_contourmap_ticks(self, y):
        if not self.gtk:
            return
        c_min = None
        c_max = None
        
        if self.current_contourmap_method:
            (c_min, c_max, lvls) = \
                    self.current_contourmap_method.get_contourmap_info()

        if ((c_max is not None) and (c_min is not None)):
            if y is not None:
                real_y = y + c_min
                # need to know which level am I in.
                level = None
                for i in range(len(lvls)):
                    if real_y <= lvls[i]:
                        level = i-1
                        if i==len(lvls):
                            i -= 1
                        break
                if level is not None:
                    # HERE 
                    self.contourmapdata.canvas_ticklayer.clear()
                    # self.contourmapdata.canvas_ticklayer.destroy()
                    # self.contourmapdata.canvas_ticklayer = self.contourmapdata.canvas.newLayer()
                    self.contourmapdata.canvas_ticklayer.raise_to_top()
                    self.contourmapdata.canvas_ticklayer.make_current()
                
                    self.contourmapdata.device.set_lineColor(
                        self.settings.contourmap_markercolor)
                    self.contourmapdata.device.set_lineWidth(
                        self.settings.contourmap_markersize)
                    width=(c_max-c_min)/self.settings.aspectratio

                    polygon = primitives.Polygon(
                        [primitives.Point(0.0, lvls[level]-c_min),
                         primitives.Point(width, lvls[level]-c_min),
                         primitives.Point(width, lvls[level+1]-c_min),
                         primitives.Point(0.0, lvls[level+1]-c_min)
                         ])
                    self.contourmapdata.device.draw_polygon(polygon)

                    self.contourmapdata.mark_value = y
                    minimum = lvls[level]
                    maximum = lvls[level] + (lvls[1]-lvls[0])
                    mainthread.runBlock(
                        self.contourlevel_min.set_text,
                        ("%.2e" % minimum,) )
                    mainthread.runBlock(
                        self.contourlevel_max.set_text,
                        ("%.2e" % maximum,) )

        # If any of the above conditions were not met, clear the
        # tick layer.
        if (c_max is None) or (c_min is None) or \
               (y is None) or (level is None):
            self.contourmapdata.canvas_ticklayer.clear()
            mainthread.runBlock(
                self.contourlevel_min.set_text, ('',) )
            mainthread.runBlock(
                self.contourlevel_max.set_text, ('',) )

        self.contourmapdata.device.show()

    # Callback for size changes of the pane containing the contourmap.
    def contourmap_config(self, x, y, width, height):
        debug.mainthreadTest()
        if self.current_contourmap_method:
            (c_min, c_max, c_lvls) = \
                    self.current_contourmap_method.get_contourmap_info()
            if c_max!=c_min:
                ppu = 1.0*height/(c_max-c_min)
                # ppu = 1.0*height/1000.0
            else:
                ppu = height # Arbitrary

            # ppu changes here can't be large, so we don't have to
            # worry about order of operations (see comment in
            # show_contourmap_info, above).
            self.contourmapdata.canvas.set_pixels_per_unit(ppu)
            self.contourmapdata.canvas.set_scrollregion(
                self.contourmapdata.canvas.get_bounds())

    def contourmap_mouse(self, event, x, y, shift, ctrl):
        debug.mainthreadTest()
        if event=="down":
            self.contourmapdata.mouse_down = 1
        elif (event=="move") and self.contourmapdata.mouse_down==1:
            subthread.execute(self.show_contourmap_ticks, (y,))
        elif event=="up":
            self.contourmapdata.mouse_down = None
            subthread.execute(self.show_contourmap_ticks, (y,))



    # Button callback.
    def contourmap_clear_marker(self, gtk):
        self.contourmapdata.mark_value = None
        self.contourlevel_min.set_text('')
        self.contourlevel_max.set_text('')
        subthread.execute(self.show_contourmap_info)


    # Overridden menu callbacks for the contourmap show/hide menu item.
    def hideLayerContourmap(self, menuitem, n):
        self.acquireGfxLock()
        try:
            ghostgfxwindow.GhostGfxWindow.hideLayerContourmap(self, menuitem, n)
            self.layerListRowChanged(n)
            self.show_contourmap_info()
        finally:
            self.releaseGfxLock()
        
    def showLayerContourmap(self, menuitem, n):
        self.acquireGfxLock()
        try:
            ghostgfxwindow.GhostGfxWindow.showLayerContourmap(self, menuitem, n)
            self.layerListRowChanged(n)
            self.show_contourmap_info()
        finally:
            self.releaseGfxLock()

    
    # GUI callbacks -- required because the GUI menu item must
    # operate on the current layer.
    def hideLayerContourmap_gui(self, menuitem):
        if self.selectedLayer is None:
            reporter.report(
                "Unable to hide contour map, no layer is selected.")
        else:
            subthread.execute(
                self.hideLayerContourmap, (menuitem,
                                           self.layerID(self.selectedLayer)) )

    def showLayerContourmap_gui(self, menuitem):
        if self.selectedLayer is None:
            reporter.report(
                "Unable to show contour map, no layer is selected.")
        else:
            subthread.execute(
                self.showLayerContourmap, (menuitem,
                                           self.layerID(self.selectedLayer)) )

    ###########################################
   
    def realizeCB(self, *args):         # gtk callback
        if not self.realized:
            self.realized = 1
            # Asynchronously zoom.
            if not self.zoomed:
                subthread.execute(self.zoomFillWindow)
        return False

        
    def close(self, *args):             # OOFMenu callback
        # The subwindow menu-removal can't depend on the existence of
        # .gtk, and it's done in the non-GUI parent, so call it
        # if this is the first time through.
        if not self.closed:
            ghostgfxwindow.GhostGfxWindow.close(self, *args)
            self.closed = 1
        if self.gtk:
            mainthread.runBlock(self.gtk.destroy) # calls destroyCB via gtk

    def destroyCB(self, *args):         # gtk callback
        # See comment in GhostGfxWindow.close about the order of operations.
        if self.gtk:
            ## tell all the miniThreads to stop and go home.
            self.device.destroy()
            self.contourmapdata.device.destroy()
            
            for tbgui in self.toolboxGUIs:
                if tbgui.active:
                    tbgui.deactivate()
                tbgui.close()
            del self.toolboxGUIs
            del self.mouseHandler
            if self.menu:
                self.menu.gui_callback=None

            self.gtk = None             # make sure this isn't repeated
            if self.menu:
                self.menu.File.Close()    # calls self.close()

##################################################

    def switchToolbox(self, chooser, tbname): # toolboxchooser callback
        self.selectToolbox(tbname)

    def selectToolbox(self, tbname):
        debug.mainthreadTest()
        if not (self.current_toolbox and (self.current_toolbox.name()==tbname)):
            self.removeMouseHandler()
            if self.current_toolbox:
                self.current_toolbox.deactivate()
            self.current_toolbox = self.getToolboxGUIByName(tbname)
            self.current_toolbox.activate()
            self.toolboxbody.foreach(self.toolboxbody.remove)
            self.toolboxbody.add(self.current_toolbox.gtk)
            self.toolboxbody.show_all()

    def getToolboxGUIByName(self, name):
        for tbgui in self.toolboxGUIs:
            if tbgui.name() == name:
                return tbgui

    def makeToolboxGUI(self, tb):
        tbgui = mainthread.runBlock(tb.makeGUI)
        if tbgui:
            self.toolboxGUIs.append(tbgui)
            self.toolboxGUIs.sort()
            self.updateToolboxChooser()
               
    def updateToolboxChooser(self):
        self.toolboxGUIs.sort()
        self.toolboxchooser.update([tb.name() for tb in self.toolboxGUIs])
                
    #############################################

    def saveImage_gui(self, menuitem):
        filename = fileselector.fileSelector(
            title='Save Image', ident='save image', mode="w")
        if filename:
            mode = fileselector.get_file_mode(filename, no_append=1)
            if mode:
                menuitem(filename=filename) # calls GhostGfxWindow.saveImage

    def saveContourmap_gui(self, menuitem):
        filename = fileselector.fileSelector(
            title='Save Contourmap', ident='save contourmap', mode="w")
        if filename:
            mode = fileselector.get_file_mode(filename, no_append=1)
            if mode:
                menuitem(filename=filename)

    # Called on a subthread, not on the main thread.
    # Argument is actually used at gfxwindow-open-time by the menu command.
    def draw(self,zoom=False):         # switchboard "redraw" callback
##        debug.fmsg()
        if self.closed:
            return
        if self.oofcanvas.is_empty():
            # *Always* zoom to fill the window on the first non-trivial draw
            zoom = True
            # Some drawing operations (canvasdot and canvastriangle in
            # particular) require pixels_per_unit to be set *before*
            # they're drawn.  This is a problem for the initial
            # drawing, so here we just guess at the best value, and
            # set ppu.  Since the canvas will be zoomed soon, it
            # doesn't matter if the value isn't quite correct.
            topwho = self.topwho("Microstructure", "Image",
                                 "Skeleton", "Mesh")
            if topwho:
                topms = topwho.getMicrostructure()
                width, height = topms.size()
                bbox = geometry.CRectangle(coord.Coord(0.0, 0.0),
                                           coord.Coord(width, height))
                mainthread.runBlock(self.zoom_bbox, (bbox,))
                
        self.acquireGfxLock()
##        debug.fmsg("Got gfx lock.")
        try:
##            debug.fmsg("Calling display.draw on ", self.display)
            self.display.draw(self, self.device)
##            debug.fmsg("Back from display.draw.")
            if zoom and self.realized:
                # We're using buffered output, the canvas may not have
                # received the data yet.  We have to force the data to
                # the canvas, and we have to wait until it's done
                # before zooming.
##                debug.fmsg("Flushing device.")
                self.device.flush_wait()
##                debug.fmsg("Flushed, zooming.")
                self.zoomFillWindow(lock=False)
##                debug.fmsg("Zoomed.")
        finally:
##            debug.fmsg("Releasing gfxlock.")
            self.releaseGfxLock()
        # Update the contourmap info, now that the appropriate layer
        # has completed its draw.
        self.show_contourmap_info()
##        debug.fmsg("Exiting.")
            
    def toggleAntialias(self, menuitem, antialias):
        ghostgfxwindow.GhostGfxWindow.toggleAntialias(
            self, menuitem, antialias)
        self.newCanvas()
        # Draw is subthread-safe, devices are all buffered.
        self.draw()

    def zoomFactor(self):
        return self.settings.zoomfactor

    def zoomIn(self, *args):
        self.acquireGfxLock()
        try:
            if self.oofcanvas and not self.oofcanvas.is_empty():
                mainthread.runBlock(self.oofcanvas.zoom,
                                    (self.settings.zoomfactor,))
                mainthread.runBlock(self.fix_step_increment)
                self.zoomed = 1
        finally:
            self.releaseGfxLock()

    def zoomInFocussed(self, menuitem, focus):
        self.acquireGfxLock()
        try:
            if self.oofcanvas and not self.oofcanvas.is_empty():
                mainthread.runBlock(self.oofcanvas.zoomAbout,
                                    (self.settings.zoomfactor, focus))
                mainthread.runBlock(self.fix_step_increment)
                self.zoomed = 1
        finally:
            self.releaseGfxLock()

    def zoomOut(self, *args):
        self.acquireGfxLock()
        try:
            if self.oofcanvas and not self.oofcanvas.is_empty():
                mainthread.runBlock(self.oofcanvas.zoom,
                                    (1./self.settings.zoomfactor,))
                mainthread.runBlock(self.fix_step_increment)
                self.zoomed = 1
        finally:
            self.releaseGfxLock()
        hadj = self.hScrollbar.get_adjustment()

    def zoomOutFocussed(self, menuitem, focus):
        self.acquireGfxLock()
        try:
            if self.oofcanvas and not self.oofcanvas.is_empty():
                mainthread.runBlock(self.oofcanvas.zoomAbout,
                                    (1./self.settings.zoomfactor, focus))
                mainthread.runBlock(self.fix_step_increment)
                self.zoomed = 1
        finally:
            self.releaseGfxLock()

    def zoomFillWindow(self, *args, **kwargs):
        # This has args because it's used as a menuitem callback.  The
        # 'lock' argument is not a menu argument, though.  It's only
        # used when zoomFillWindow is called explicitly from within
        # another drawing routine, to indicate that the gfxLock has
        # already been acquired.
        lock = kwargs.get('lock', True) # default is to acquire the lock
        if lock:
            self.acquireGfxLock()
        try:
            mainthread.runBlock(self.zoomFillWindow_thread)
        finally:
            if lock:
                self.releaseGfxLock()


    def zoomFillWindow_thread(self):
        debug.mainthreadTest()
        if self.closed:
            return
        if self.oofcanvas and not self.oofcanvas.is_empty():
            bbox = self.oofcanvas.get_bounds()
            self.zoom_bbox(bbox)

    def zoom_bbox(self, bbox):
        debug.mainthreadTest()
        width = self.oofcanvas.get_width_in_pixels()-1
        height = self.oofcanvas.get_height_in_pixels()-1
        if bbox.width() > 0 and bbox.height() > 0:
            xf = width/bbox.width()
            yf = height/bbox.height()
            ppu = min(xf, yf)       # pixels per unit
            oldppu = self.oofcanvas.get_pixels_per_unit()
            if ppu > oldppu:    # See comment in show_contourmap_info().
                self.oofcanvas.set_scrollregion(bbox)
                self.oofcanvas.set_pixels_per_unit(ppu)
            else:
                self.oofcanvas.set_pixels_per_unit(ppu)
                self.oofcanvas.set_scrollregion(bbox)
            self.zoomed = 1
        self.fix_step_increment()

    def fix_step_increment(self):
        # For some reason, the step_increment on the canvas scroll
        # bars is zero, so the scroll arrows don't work.  This fixes
        # that.
        debug.mainthreadTest()
        hadj = self.hScrollbar.get_adjustment()
        if hadj.step_increment == 0.0:
            hadj.step_increment = hadj.page_increment/16
        vadj = self.vScrollbar.get_adjustment()
        if vadj.step_increment == 0.0:
            vadj.step_increment = vadj.page_increment/16

    def set_zoom(self, scrollreg, ppu):
        mainthread.runBlock(self.set_zoom_thread, (scrollreg, ppu))
    def set_zoom_thread(self, scrollreg, ppu):
        self.acquireGfxLock()
        try:
            oldppu = self.oofcanvas.get_pixels_per_unit()
            if ppu > oldppu:            # See comment in show_contourmap_info().
                self.oofcanvas.set_scrollregion(scrollreg)
                self.oofcanvas.set_pixels_per_unit(ppu)
            else:
                self.oofcanvas.set_pixels_per_unit(ppu)
                self.oofcanvas.set_scrollregion(scrollreg)
            self.zoomed = 1
        finally:
            self.releaseGfxLock()

    # GUI override of menu callback for new contourmap aspect ratio.
    # GUI callbacks are required because, when the settings change,
    # you have to redraw the window.
    def aspectRatio(self, menuitem, ratio):
        self.acquireGfxLock()
        try:
            self.settings.aspectratio = ratio
        finally:
            self.releaseGfxLock()
        self.show_contourmap_info()

    # Overridden menu callbacks for these settings.
    def contourmapMarkSize(self, menuitem, width):
        self.acquireGfxLock()
        try:
            self.settings.contourmap_markersize = width
            v = self.contourmapdata.mark_value
            self.contourmapdata.mark_value = None
        finally:
            self.releaseGfxLock()
        self.show_contourmap_ticks(v)

    def contourmapMarkColor(self, menuitem, color):
        self.acquireGfxLock()
        try:
            self.settings.contourmap_markercolor = color
            v = self.contourmapdata.mark_value
            self.contourmapdata.mark_value = None
        finally:
            self.releaseGfxLock()
        self.show_contourmap_ticks(v)

    def toggleLongLayerNames(self, menuitem, longlayernames):
        self.acquireGfxLock()
        try:
            self.settings.longlayernames = longlayernames
            self.fillLayerList()
        finally:
            self.releaseGfxLock()
        
    # Sets background color for both canvases.
    def bgColor(self, menuitem, color):          # OOFMenu callback
        self.acquireGfxLock()
        try:
            ghostgfxwindow.GhostGfxWindow.bgColor(self, menuitem, color)
            mainthread.runBlock(self.oofcanvas.set_bgColor, (color,))
            mainthread.runBlock(self.contourmapdata.canvas.set_bgColor,
                                (color,))
        finally:
            self.releaseGfxLock()

    def marginCB(self, menuitem, fraction):
        self.acquireGfxLock()
        try:
            self.settings.margin = fraction
            mainthread.runBlock(self.oofcanvas.set_margin, (fraction,))
        finally:
            self.releaseGfxLock()
        
    ################################

    def newLayer_gui(self, menuitem):
        layereditor.menu.LayerSet.New(window=self.name)

    def editLayer_gui(self, menuitem):
        layereditor.menu.LayerSet.Edit(
            window=self.name, layer_number=self.layerID(self.selectedLayer))

    def newLayer(self, displaylayer):
        self.fillLayerList()
        
    def deleteLayer_gui(self, menuitem):
        if self.selectedLayer is None:
            top().message_out('No layer is selected!\n')
        else:
            self.suppressSelectionSignals()
            try:
                self.menu.Layer.Delete(n=self.layerID(self.selectedLayer))
            finally:
                self.allowSelectionSignals()

    # This is an override of the command-line menu callback in
    # ghostgfxwindow.  It exists in addition to the GUI callback
    # below, which is also required because it operates on the
    # current layer.
    def hideLayer(self, menuitem, n):   # OOFMenu callback
        ghostgfxwindow.GhostGfxWindow.hideLayer(self, menuitem, n)
        mainthread.runBlock(self.hideLayer_thread, (menuitem, n))
    def hideLayer_thread(self, menuitem, n):
        self.display.hide_layer(self.device, n) # hide layer in canvas
        self.layerListRowChanged(n)
        # Update the contourmap.
        self.contourmap_newlayers()
        subthread.execute(self.show_contourmap_info)

        
    def hideLayer_gui(self, menuitem):  # OOFMenu GUI callback.
        if self.selectedLayer is None:
            reporter.report('No layer is selected!')
        else:
            self.menu.Layer.Hide(n=self.layerID(self.selectedLayer))

        
    def showLayer(self, menuitem, n):   # OOFMenu callback.
        ghostgfxwindow.GhostGfxWindow.showLayer(self, menuitem, n)
        mainthread.runBlock(self.showLayer_thread, (menuitem, n))
    def showLayer_thread(self, menuitem, n):
        self.display.show_layer(self.device, n) # show layer in canvas
        self.layerListRowChanged(n)
        # Update the contourmap.
        self.contourmap_newlayers()
        subthread.execute(self.show_contourmap_info)


    def showLayer_gui(self, menuitem):  # OOFMenu GUI callback
        if self.selectedLayer is None:\
            reporter.report('No layer is selected!')
        else:
            self.menu.Layer.Show(n=self.layerID(self.selectedLayer))

    def selectLayer(self, n):
        if self.selectedLayer is not None:
            self.deselectLayer(self.selectedLayer)
        ghostgfxwindow.GhostGfxWindow.selectLayer(self, n)
        mainthread.runBlock(self.selectLayer_thread, (n,))
    def selectLayer_thread(self, n):
        self.suppressSelectionSignals()
        self.layerListView.get_selection().select_path(len(self.display)-n-1)
        self.allowSelectionSignals()
    def deselectLayer(self, n):
        if self.selectedLayer is not None:
            mainthread.runBlock(self.deselectLayer_thread, (n,))
            ghostgfxwindow.GhostGfxWindow.deselectLayer(self, n)
    def deselectLayer_thread(self, n):
        self.suppressSelectionSignals()
        self.layerListView.get_selection().unselect_all()
        self.allowSelectionSignals()
    def deselectAll(self):
        # called by createDefaultLayers and sb 'deselect all gfxlayers'
        if self.selectedLayer is not None:
            mainthread.runBlock(self.deselectAll_thread)
    def deselectAll_thread(self):
        self.suppressSelectionSignals()
        self.layerListView.get_selection().unselect_all()
        ghostgfxwindow.GhostGfxWindow.deselectAll(self)
        self.allowSelectionSignals()
        
    # At layer-removal time, it's necessary to explicitly redraw the
    # contourmap, because the main canvas layers have not been redrawn,
    # so the usual automatic redraw has not taken place.
    def removeLayer(self, layer):
        ghostgfxwindow.GhostGfxWindow.removeLayer(self, layer)
        self.show_contourmap_info()
    
    def raiseLayer_gui(self, menuitem):
        if self.selectedLayer is None:
            top().message_out('No layer is selected!\n')
        else:
            self.menu.Layer.Raise.One_Level(
                n=self.layerID(self.selectedLayer))
    def raiseToTop_gui(self, menuitem):
        if self.selectedLayer is None:
            top().message_out('No layer is selected!\n')
        else:
            self.menu.Layer.Raise.To_Top(
                n=self.layerID(self.selectedLayer))
    def lowerLayer_gui(self, menuitem):
        if self.selectedLayer is None:
            top().message_out('No layer is selected!\n')
        else:
            self.menu.Layer.Lower.One_Level(
                n=self.layerID(self.selectedLayer))
    def lowerToBottom_gui(self, menuitem):
        if self.selectedLayer is None:
            top().message_out('No layer is selected!\n')
        else:
            self.menu.Layer.Lower.To_Bottom(
                n=self.layerID(self.selectedLayer))
    def raiseLayerByGUI(self, layer):
        self.menu.Layer.Raise.One_Level(n=self.layerID(layer))
    def lowerLayerByGUI(self, layer):
        self.menu.Layer.Lower.One_Level(n=self.layerID(layer))
        
    ##############################################

    # def cloneWindow(self, *args):
        ## This doesn't work completely correctly as far as copying sizes goes.
        # Create the clone
    #     clone = ghostgfxwindow.GhostGfxWindow.cloneWindow(self, *args)
##        # Tell the clone what its size should be.
##        gtkutils.copyWidgetSizes(self.gtk, clone.gtk)
##        if not self.oofcanvas.is_empty():
##            sr = self.oofcanvas.get_scrollregion()
##            ppu = self.oofcanvas.get_pixels_per_unit()
##            clone.set_zoom(sr, ppu)
##            clone.oofcanvas.hScrollTo(self.oofcanvas.hScrollPosition())
##            clone.oofcanvas.vScrollTo(self.oofcanvas.vScrollPosition())
    #     return clone


    ##### Scrolling ###########

    def scrlReleaseCB(self, scrollbar, event, which):
        if which == 'h':
            self.menu.Settings.Scroll.Horizontal(
                position=scrollbar.get_adjustment().get_value())
        else:
            self.menu.Settings.Scroll.Vertical(
                position=scrollbar.get_adjustment().get_value())

    def hScrollCB(self, menuitem, position): # OOFMenu callback
        ghostgfxwindow.GhostGfxWindow.hScrollCB(self, menuitem, position)
        mainthread.runBlock(
            self.hScrollbar.get_adjustment().set_value, (position,))

    def vScrollCB(self, menuitem, position): # OOFMenu callback
        ghostgfxwindow.GhostGfxWindow.vScrollCB(self, menuitem, position)
        mainthread.runBlock(
            self.vScrollbar.get_adjustment().set_value, (position,))

    def hScrollPosition(self):
        return self.hScrollbar.get_adjustment().value
    def vScrollPosition(self):
        return self.vScrollbar.get_adjustment().value

    #### Mouse clicks ############

    def setMouseHandler(self, handler):
        self.mouseHandler = handler

    def removeMouseHandler(self):
        self.mouseHandler = mousehandler.nullHandler

    def mouseCB(self, eventtype, x, y, shift, ctrl):
        debug.mainthreadTest()
        global _during_callback
        _during_callback = 1
        if self.mouseHandler.acceptEvent(eventtype):
            if eventtype == 'up':
                self.mouseHandler.up(x,y, shift, ctrl)
            elif eventtype == 'down':
                self.mouseHandler.down(x,y, shift, ctrl)
            elif eventtype == 'move':
                self.mouseHandler.move(x,y, shift, ctrl)
        _during_callback = 0

    def setRubberband(self, rubberband):
        self.rubberband = rubberband
        self.oofcanvas.set_rubberband(rubberband)

###########################################

## Support for logging and replaying mouse clicks.

# Although the Canvas *is* a gtk Widget, it's not a pygtk Widget, and
# it's hard to use the widget logging machinery directly on it. So we
# pretend that it's some other kind of gtk object and use the
# adoptGObject machinery instead.  It's adopted by
# GfxWindow.canvasTable.  adoptGObject is told the name of the window.
# The log uses the window name and findCanvasRoot to retrieve the
# Canvas's root, which is the crucial bit for emitting signals.

def findCanvasRoot(gtkobj, windowname):

    # In gui logs, returns the root object of the canvas of the given
    # gfxwindow.  Put into the logs by AdopteeLogger.location, via
    # CanvasLogger.location.  See GfxWindow.newCanvas().
    
    window = gfxmanager.gfxManager.getWindow(windowname)
    return window.oofcanvas.rootitem()

def findCanvasGdkWindow(windowname):
    window = gfxmanager.gfxManager.getWindow(windowname)
    return window.oofcanvas.widget().window

# desired_events is a list of all of the event types that should be
# logged on the canvas.

desired_events = [gtk.gdk.BUTTON_PRESS,
                  gtk.gdk.BUTTON_RELEASE,
                  gtk.gdk.MOTION_NOTIFY]

_canvaslogging_initialized = False

def init_canvas_logging(canvasgroup):
    global _canvaslogging_initialized
    if _canvaslogging_initialized:
        return
    _canvaslogging_initialized = True

    # Inject findCanvasRoot into the gtklogger namespace, which is
    # where gui scripts are run.
    gtklogger.findCanvasRoot = findCanvasRoot
    gtklogger.findCanvasGdkWindow = findCanvasGdkWindow

    # The GtkLogger for Canvas events has to be defined *after* the
    # first canvas has been created, because the GnomeCanvasGroup
    # class isn't in pygtk.  We snag the object returned by
    # OOFCanvas::rootitem and use its class for the CanvasLogger.
    class CanvasLogger(gtklogger.AdopteeLogger):
        classes = (canvasgroup.__class__,)
        def location(self, object, *args):
            self.windowname = object.oofparent_access_kwargs['windowname']
            return super(CanvasLogger, self).location(object, *args)
        buttonup = True
        def record(self, object, signal, *args):
            if signal == "event":
                event = args[0]
                if event.type in desired_events:
                    # The following if/elif block is not really in the
                    # spirit of gui logging, since it more or less
                    # repeats some of the code that it's supposed to
                    # be testing.  But the logs get too full of
                    # motion_notify events if they're all logged. This
                    # block removes extraneous motion notify events
                    # from the log.
                    if event.type == gtk.gdk.BUTTON_PRESS:
                        CanvasLogger.buttonup = False
                    elif event.type == gtk.gdk.BUTTON_RELEASE:
                        CanvasLogger.buttonup = True
                    elif event.type == gtk.gdk.MOTION_NOTIFY:
                        if gtklogger._suppress_motion_events and self.buttonup:
                            return self.ignore
                    # event.type.value_name is something of the form
                    # GDK_XXXXXX, but the python variable is
                    # gtk.gdk.XXXXXX, so we have to strip off the
                    # "GDK_".
                    eventname = event.type.value_name[4:]
                    return [
                        "canvasobj = %s" % self.location(object, *args),
                        "canvasobj.emit('event', event(gtk.gdk.%s,x=%20.13e,y=%20.13e,state=%d,window=findCanvasGdkWindow('%s')))"
                        % (eventname, event.x, event.y, event.state,
                           self.windowname)
##   "%s.emit('event', event(gtk.gdk.%s, x=%20.13e, y=%20.13e, state=%d))" % 
##                           (self.location(object, *args),
##                            eventname, event.x, event.y, event.state)
                        ]
                return self.ignore      # silently ignore other events
            super(CanvasLogger, self).record(object, signal, *args)

##############################################

# This function redefines the one in GfxWindowManager when the GUI
# code is loaded.

def _newWindow(self, name, count, **kwargs):
    if top(): # if in GUI mode
        return GfxWindow(name, count, self, **kwargs)
    return ghostgfxwindow.GhostGfxWindow(name, count, self, **kwargs)

gfxmanager.GfxWindowManager._newWindow = _newWindow

