# -*- python -*-
# $RCSfile: guilogger.py,v $
# $Revision: 1.11 $
# $Author: langer $
# $Date: 2006/05/08 17:27:33 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from oof2.common import debug
from oof2.common import thread_enable
from oof2.common.IO import mainmenu
from oof2.common.IO import oofmenu
from oof2.common.IO import reporter
from oof2.common.IO import parameter
from oof2.common.IO import progressbar_delay
from oof2.common.IO.GUI import gtklogger
import gobject
import gtk
import sys
import traceback

guidebugmenu = mainmenu.debugmenu.addItem(oofmenu.OOFMenuItem("GUI_Logging",
                                                              ordering=1000,
                                                              no_log=True,
                                                              post_hook=None))

###################################

# menucheckpoint is installed as the main menu's post_hook function.
# It's run after every menu command, and inserts a checkpoint in the
# log file.

def menucheckpoint(menuitem):
    gtklogger.checkpoint(menuitem.path())

##################################

_recording = False
_replaying = False

def recording():
    return _recording

def replaying():
    return _replaying

def logFinished():                      # called by GUILogPlayer when done
    global _replaying
    _replaying = False
    mainmenu.OOF.removeOption('post_hook')
    debug.fmsg("Finished replaying gui log file.")

############################

    
def startLog(menuitem, filename):
    gtklogger.reset_checkpoints()
    menuitem.root().setOption('post_hook', menucheckpoint)
    if debug.debug():
        dblevel = 3
    else:
        dblevel = 0
    global _recording
    _recording = True
    gtklogger.start(filename, debugLevel=dblevel)

def startLogGUI(menuitem):
    from oof2.common.IO.GUI import fileselector
    file = fileselector.fileSelector(ident="guiscript", title="Save GUI Log",
                                     mode='w')
    if file is not None:
        menuitem(filename=file)

guidebugmenu.addItem(oofmenu.OOFMenuItem(
    'Record',
    callback=startLog,
    gui_callback=startLogGUI,
    params=[parameter.StringParameter('filename',
                                      'oofgui.log', tip="Name of the file.")],
    ellipsis=1,
    help="Save GUI events in a Python script",
    ))

############################

def stopLog(menuitem):
    global _recording
    _recording = False
    menuitem.root().removeOption('post_hook')
    gtklogger.stop()

guidebugmenu.addItem(oofmenu.OOFMenuItem(
    'Stop',
    callback=stopLog,
    no_log=1,
    help="Stop recording GUI events"))

############################

# loggererror is installed as the gtklogger handler for exceptions
# that occur during playback.  It's not possible to use the standard
# oof2 exception handling mechanism directly, because it doesn't
# handle the gtk thread locks properly in this case (ie, inside an
# idle callback on the main thread, but with
# oof_mainiteration._inside_idle_callback==0).

def loggererror(exc, line):
    from oof2.common.IO.GUI import reporter_GUI
    type, value, tb = sys.exc_info()
    tblist = traceback.extract_tb(tb)
    reporter_GUI.gui_printTraceBack(type, value, tblist)
    return True

####

def loadLogGUI(menuitem):
    from oof2.common.IO.GUI import fileselector
    file = fileselector.fileSelector(ident="guiscript",
                                     title="Load GUI Log", mode="w")
    if file is not None:
        menuitem(filename=file)

def loadLog(menuitem, filename):
    if gtklogger.GUILogPlayer.current is not None:
        raise ooferror.ErrUserError(
            "Multiple GUI logs cannot be replayed simultaneously!")
#    gtklogger.reset_checkpoints()
    debug.fmsg("Loading gui script", filename)
    menuitem.root().setOption('post_hook', menucheckpoint)
    dblevel = 0
    if debug.debug():
        dblevel = 3
    #dblevel = 4
    global _replaying
    _replaying = True

    # When replaying, we have to make sure that progress bars *always*
    # appear, so we set the delay time to 0.  If the delay time is
    # non-zero, then a script recorded on a slow machine would insert
    # a checkpoint for opening the activity viewer window, but a
    # faster machine might never open the window, and would wait
    # forever for the checkpoint when replaying the script.
    progressbar_delay.set_delay(None, 0)

    gobject.idle_add(gtklogger.GUILogPlayer(filename,
                                            finishCB=logFinished,
                                            debugLevel=dblevel,
                                            threaded=thread_enable.query(),
                                            exceptHook=loggererror))

guidebugmenu.addItem(oofmenu.OOFMenuItem(
    'Replay',
    callback=loadLog,
    gui_callback=loadLogGUI,
    params=[parameter.StringParameter('filename', 'oofgui.log',
                                      tip="Name of the file.")],
    ellipsis=1,
    help="Load a GUI log file"))

#####################

def sanity_check(menuitem):
    gtklogger.comprehensive_sanity_check()

guidebugmenu.addItem(oofmenu.OOFMenuItem(
    'Sanity_Check',
    callback=sanity_check,
    no_log=1,
    help="Check that widget names are unique"))

