# -*- python -*-
# $RCSfile: mainmenu.py,v $
# $Revision: 1.146 $
# $Author: langer $
# $Date: 2006/08/24 02:18:53 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

#
#  The main oof menu -- with recording capability and so forth.

# Lots of miscellaneous commands are defined in this file because they
# didn't seem to merit a file of their own.

from oof2.SWIG.common import ooferror
from oof2.SWIG.common import switchboard
from oof2.SWIG.common import threadstate
from oof2.common import debug
from oof2.common import subthread
from oof2.common import thread_enable
from oof2.common import utils
from oof2.common.IO import oofmenu
from oof2.common.IO import parameter
from oof2.common.IO import reporter
from oof2.common.IO import scriptloader
from oof2.common.IO.gfxmanager import gfxManager
from oof2.common.IO import xmlmenudump
from types import StringType, FloatType, IntType
import code
import sys
import tempfile
import atexit
import os

# Parameter = parameter.Parameter
StringParameter = parameter.StringParameter
IntParameter = parameter.IntParameter
FloatParameter = parameter.FloatParameter

OOFMenuItem = oofmenu.OOFMenuItem
OOFRootMenu = oofmenu.OOFRootMenu
CheckOOFMenuItem = oofmenu.CheckOOFMenuItem

OOF = OOFRootMenu(
    'OOF',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/common/menu/oof.xml'))

################

# Automatically log all menu commands to a temp file.  The secure way
# of getting a temp file is new in Python 2.3, so check the version
# number.

version = sys.version_info[0] + sys.version_info[1]*0.1
if version >= 2.3:
    fd, logpath = tempfile.mkstemp(prefix='oof2-', suffix='.py')
    tmplogfile = os.fdopen(fd, 'w')
else:
    logpath = tempfile.mktemp(suffix='-oof2.py')
    tmplogfile = open(logpath, 'w')

def _tmplog(s):
    if tmplogfile:
        print >> tmplogfile, s
        tmplogfile.flush()
    
OOF.addLogger(oofmenu.MenuLogger(_tmplog))

# Remove the log file automatically if the program exits cleanly
def cleanlog():
    global tmplogfile, logpath
    if tmplogfile:
        try:
            tmplogfile.close()
            os.remove(logpath)
        except:
            pass
        logpath = ""
        tmplogfile = None
atexit.register(cleanlog)


################

## File menu

_filemenu = OOF.addItem(OOFMenuItem(
    'File',
    help="Commands for saving and loading data, and quitting."))

_loadmenu = _filemenu.addItem(OOFMenuItem(
    'Load',
    help="Commands for loading datafiles and scripts.",
    discussion="<para>Commands to load datafiles and scripts.</para>"))

class PScriptLoader(scriptloader.ScriptLoader):
    # A ScriptLoader that supports a progress bar.
    def __init__(self, filename, progbar):
        self.progressbar = progbar
        scriptloader.ScriptLoader.__init__(self, filename,
                                         sys.modules['__main__'].__dict__)  
    def progress(self, current, total):
        self.progressbar.set_progress((1.0*current)/total)
        self.progressbar.set_message("%d/%d lines" % (current, total))
    def stop(self):                     # called by ScriptLoader to abort loop
        return self.progressbar.query_stop()
    def done(self):                     # called by ScriptLoader when finished
        if self.error:
            self.progressbar.set_failure()
        else:
            self.progressbar.set_success()
        scriptloader.ScriptLoader.done(self)

def loadscript(menuitem, filename):
    if filename is not None:
        debug.fmsg('reading', filename, 'in thread',
                   threadstate.findThreadNumber())
        interp = PScriptLoader(filename, menuitem.getProgressBar("continuous"))
        interp.run()
        debug.fmsg('finished reading', filename)

_loadmenu.addItem(OOFMenuItem(
    'Script',
    callback=loadscript,
    params=[StringParameter('filename', 'logfile', tip="Name of the file.")],
    no_log=1,
    ellipsis=1,
    threadable=oofmenu.THREADABLE_GUI,
    accel='l',
    help="Execute a Python script.",
    post_hook=None,
    discussion=xmlmenudump.loadFile('DISCUSSIONS/common/menu/loadscript.xml')
    ))

# "Data" files are distinguished from scripts in that they're not read
# by the Python interpreter directly, and therefore can't contain
# arbitrary Python code.  They can only contain menu commands from the
# OOF.LoadData menu.  The command arguments can be constants or
# variables and functions that are defined in the main OOF namespace,
# or lists and tuples thereof.

def loaddata(menuitem, filename):
    if filename is not None:
        from oof2.common.IO import datafile
        debug.fmsg('loading', filename)
        datafile.readDataFile(filename, OOF.LoadData,
                              menuitem.getProgressBar("continuous"))
        debug.fmsg('done loading', filename)

_loadmenu.addItem(OOFMenuItem(
    'Data',
    callback=loaddata,
    threadable=oofmenu.THREADABLE,
    params=[StringParameter('filename', tip="Name of the file.")],
    ellipsis=1,
    help="Load a data file.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/common/menu/loaddatafile.xml')
    ))

OOF.addItem(oofmenu.OOFMenuItem(
    "LoadData", secret=1,
    help="Commands used in data files.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/common/menu/loaddata.xml'),
    post_hook=None, # Don't include checkpoints in gui logs
    ))


def saveLog(menuitem, filename):
    file = open(filename, 'w')
    menuitem.root().saveLog(file)
    file.close()

_savemenu = _filemenu.addItem(
    OOFMenuItem('Save', help='Create data files and scripts.'))

_savemenu.addItem(OOFMenuItem(
    'Python_Log',
    callback=saveLog,
    ordering=10,
    params=[StringParameter('filename', tip="Name of the file.")],
    accel='s',
    ellipsis=1,
    no_log=1,
    help="Save the current session as a Python script.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/common/menu/savelog.xml')
    ))

#################################


def quitCmd(menuitem):
    from oof2.common import quit
    quit.quit()

# TODO: THE FOLLOWING COMMENT IS OBSOLETE, IT IS HOPED.
# TODO LATER:

# On Linux, there's a problem with running the Quit menu item from
# scripts -- it doesn't quit properly, it just hangs, with the main
# thread stuck in "rt_sigsuspend".  There's no obvious lock problem,
# and there's somewhat dated indirect evidence that this may be a
# problem with the Linux kernel, or possibly with the Linux threading
# model.

# This appears to happen to the Quit command in both GUI mode and text
# mode.

# The difficulty seems to involve order-of-operations problems in
# shutting down threads.  When Quit is run from a script, there are
# many threads involved -- the thread interpreter itself runs on a
# "mini-thread" started by subthread.execute from the GUI start
# process.  All the menu items run from this thread detect that they
# are not being started from the main thread, and are run subthreads,
# with their callers blocking on a "join" call.  The callback of the
# Quit menu item itself is run on the main thread by means of an idle
# callback in the graphics event loop.  It is this code which runs the
# thread shut-down code, and which kills the other threads.  The
# script-running mini-thread seems to be the one whose shutdown causes
# problems, and there's no obvious simple solution -- you can't make
# script-running minithreads immortal, or scripts become
# uninterruptible.  It's hard to find out which mini-thread spawned
# the eventual Quit command, because there are two intervening thread
# transitions, and this is true for non-Quit menu-items also.  A
# solution may be to keep track of the chain of "calling" threads, so
# the Quit command can determine if it has a minithread ancestor, and
# "immortalize" it.  This has not yet been attempted.

# The importance of this problem is debatable.  Running a script in
# GUI mode which contains a Quit command is a odd thing to want to do
# -- the purpose of the GUI is to allow the user to interact with
# objects in the program, and a script which does all this and then
# exits wouldn't allow such interaction, so the point of the GUI is a
# little unclear.  The Quit command is similarly redundant in text
# mode -- OOF can be run non-interactively in text mode via the
# "--batch" option, with which scripts will simply exit when they get
# to the end.  This can also run multiple scripts, and is more
# flexible than including explicit "Quit" commands.



_filemenu.addItem(OOFMenuItem(
    'Quit',
    callback=quitCmd,
    accel='q',
    help="Don't give up so easily!",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/common/menu/quit.xml'),
    threadable = oofmenu.UNTHREADABLE,
    no_log=1
    )) 


##def dwiw(*args):
##    print "I'm sorry, Dave, I can't do that."
##_filemenu.addItem(OOFMenuItem('Do_What_I_Want', callback=dwiw))


##################################

settingsmenu = OOF.addItem(OOFMenuItem(
    'Settings',
    help="Global settings",
    discussion="""
    <para>
    Commands for setting parameters that don't belong anywhere else.
    </para>"""))

fontmenu = settingsmenu.addItem(OOFMenuItem(
    "Fonts",
    help="Set fonts used in the GUI."
    ))

def setFont(menuitem, fontname):
    switchboard.notify('change font', fontname)

fontmenu.addItem(OOFMenuItem(
    "Widgets",
    callback=setFont,
    params=[parameter.StringParameter('fontname', tip="The name of a font.")],
    help="Set the font to use for labels, menus, buttons, etc. in the graphical user interface.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/common/menu/widgetfont.xml')
    ))

def setFixedFont(menuitem, fontname):
    switchboard.notify('change fixed font', fontname)

fontmenu.addItem(OOFMenuItem(
    "Fixed",
    callback=setFixedFont,
    params=[parameter.StringParameter('fontname', tip='The name of a font.')],
    help="Set the fixed-width font to use in text displays.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/common/menu/textfont.xml')
    ))

def setTheme(menuitem, theme):
    switchboard.notify('change theme', theme)

settingsmenu.addItem(OOFMenuItem(
    "Theme",
    callback=setTheme,
    params=[parameter.StringParameter('theme',
                                      tip="The name of a gnome theme.")],
    help="Set the look and feel of the graphical user interface.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/common/menu/theme.xml')
    ))

bufsizemenu = settingsmenu.addItem(OOFMenuItem(
    "UndoBuffer_Size",
    help="Set the size of history buffers"))

##################################

## Subwindows menu.

_windowmenu = OOFMenuItem(
    'Windows',
    help="Menus for opening and raising windows.")
OOF.addItem(_windowmenu)

def dummy(menuitem): pass   # Dummy callback, so trivial menu items get logged.

# Add an entry for the main window, which, when clicked, raises it.
_windowmenu.addItem(OOFMenuItem(
    "OOF2",
    callback=dummy,
    help="Raise the main OOF2 window.",
    discussion="""<para>
    Every &oof2; subwindow has this menu available. It's a good way to
    locate the main &oof2; window, if it's out of sight.
    </para>"""))

# The Console is a no-op in text mode.

def consolation(menuitem):
    print "There, there, I'm sure everything will be fine."

_windowmenu.addItem(OOFMenuItem(
    'Console',
    callback=consolation,
    help="Open or raise the Python console interface.",
    no_log=1,
    discussion="""<para>
    The &oof2; <link linkend='Section:Windows:Console'>Console</link>
    provides a way of executing arbitrary Python code while running
    &oof2; in graphics mode.
    </para>"""))

##################

_graphicsmenu = OOFMenuItem(
    'Graphics',
    help="Graphical display of &oof2; objects.",
    discussion="""<para>
    Graphics windows are discussed in <xref
    linkend='Chapter:Graphics'/>.
    </para>""" )

def openGfx(menuitem):
    window = gfxManager.openWindow()
    window.draw(zoom=True)
    
_graphicsmenu.addItem(OOFMenuItem("New",
                                  callback=openGfx,
                                  help="Open a new graphics window.",
                                  accel='g',
                                  discussion="<para>Create a new <link linkend='Chapter:Graphics'>Graphics Window</link>.</para>"))

_windowmenu.addItem(_graphicsmenu)

_windowmenu.addItem(OOFMenuItem(
    'Layer_Editor',
    callback=dummy,
    help="Open or raise the Layer Editor window.",
    discussion="""<para>
Raise the <link linkend='Section:Graphics:LayerEditor'>Layer
Editor</link> window, if it has already been opened. If not, open it.
</para>"""
    ))

_windowmenu.addItem(OOFMenuItem(
    'Activity_Viewer',
    callback=dummy,
    help="Open or raise the Activity Viewer window.",
    discussion="""<para>
Raise the <link linkend='Section:Windows:ActivityViewer'>Activity
Viewer</link> window, if it is open. If not, open it.
</para>"""
    ))

#################################

## Help menu

helpmenu = OOF.addItem(OOFMenuItem('Help', help_menu=1))

debugmenu = helpmenu.addItem(OOFMenuItem(
    'Debug',
    discussion=
    """
<para>Tools for figuring out what's going on when it's not going well.
Mostly of interest to the developers.</para>
"""
    ))

def set_debug(menuitem, state):
    if state:
        debug.set_debug_mode()
    else:
        debug.clear_debug_mode()
        
debugmenu.addItem(CheckOOFMenuItem(
    'Debug',
    debug.debug(),
    callback=set_debug,
    help='Turn debugging mode on and off.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/common/menu/debug.xml')
    ))

debugmenu.addItem(CheckOOFMenuItem(
    'Verbose_Switchboard',
    switchboard.switchboard.verbose,
    callback=switchboard.verbose,
    help='Print all switchboard calls as they occur.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/common/menu/verbosesb.xml')
    ))

def setWarnPopups(menuitem, value):
    reporter.messagemanager.set_warning_pop_up(value)

helpmenu.addItem(CheckOOFMenuItem(
    'Popup_warnings',1,
    ordering=-8,
    callback=setWarnPopups,
    help="Display warnings in a pop-up window or just in the message window?",
    discussion="""<para>
    If <command>Popup_warnings</command> is true, warning messages
    will appear in an annoying pop-up window.  If it's false, they'll
    appear only in the <link
    linkend='Section:Windows:Messages'>Messages</link> window.
    </para>"""))

####################################

## Profiling functions, in the Debug menu

profmenu = debugmenu.addItem(OOFMenuItem('Profile'))
prof = None

def profile_start(menuitem, filename, fudge):
    global prof
##    import hotshot
##    prof = hotshot.Profile(filename)
##    prof.start()
    debug.fmsg()
    if thread_enable.enabled():
        reporter.warn("Multithreaded profiling is unreliable!\nUse the --unthreaded startup option.")
    from oof2.common.EXTRA import profiler
    prof = profiler.Profiler(filename, fudge=fudge)

def profile_stop(menuitem):
    global prof
    prof.stop()
##    prof.close()
    prof = None
    
profmenu.addItem(OOFMenuItem(
    'Start',
    callback=profile_start,
    threadable = oofmenu.UNTHREADABLE,
    params=[StringParameter('filename', 'prof.out', tip="File name."),
            FloatParameter('fudge', 2.89e-6, tip="Fudge factor.")],
    help="Begin measuring execution time.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/common/menu/profileStart.xml')
    ))
    
profmenu.addItem(OOFMenuItem(
    'Stop',
    callback=profile_stop,
    threadable = oofmenu.UNTHREADABLE,
    help="Stop measuring execution time.",
    discussion="<para>Stop profiling, and save the data in the file specified in <xref linkend='MenuItem:OOF.Help.Debug.Profile.Start'/>.</para>"
                             ))

def proffudge(menuitem, iterations):
    from oof2.common.EXTRA import profiler
    fudge = profiler.calibrate_profiler(iterations)
    helpmenu.Debug.Profile.Start.get_arg('fudge').value = fudge
    reporter.report('fudge =', fudge)

profmenu.addItem(OOFMenuItem(
    'FudgeFinder',
    callback=proffudge,
    threadable = oofmenu.UNTHREADABLE,
    params=[IntParameter('iterations', 1000, tip="Number of iterations.")],
    help='Find the factor to compensate for time spent in the profiler itself.',
    discussion="""
    <para>Find the machine dependent fudge factor that the profiler
    uses to compensate for function calling overhead in the profiler
    itself, by measuring how long it takes to call the profiler
    <varname>iterations</varname> times.  </para>
    """
    ))




####################################

## The following functions are included only if --debug is provided on
## the command line at start up.  Users shouldn't be interested in
## them.

if debug.debug():
    errmenu = helpmenu.addItem(OOFMenuItem('Error', no_doc=1))


    def _warning(menuitem):
        debug.fmsg()
        def _subwarning():
            debug.fmsg()
            reporter.warn("You'd better be home by 11, young lady!")
        subthread.execute(_subwarning)

    errmenu.addItem(OOFMenuItem('Warning', callback=_warning,
                                help='Actual numbers may vary.'))

    from oof2.SWIG.common import cdebug
    def _segfault(menuitem, delay):
        cdebug.segfault(delay)

    errmenu.addItem(OOFMenuItem('SegFault', callback=_segfault,
                                 threadable=oofmenu.THREADABLE,
                                 params=[IntParameter('delay', 10,
                                                      tip="Delay time.")],
                                 help='For external use only.  Slippery when wet.'))

    def _pyerror(menuitem):
        debug.fmsg()
        if reporter.query("Really?", "Yes", "No!") == "Yes":
            # os.abort()
            # raise ooferror.ErrPyProgrammingError("Blam!")
            reporter.warn("Look out!  Behind you!")
            raise AttributeError("Oops!")

    errmenu.addItem(OOFMenuItem('PyError', callback=_pyerror,
                                 threadable=oofmenu.THREADABLE,
                                 help='Do not taunt PyError.'))

    def _cerror(menuitem):
        debug.fmsg()
        if reporter.query("Really?", "Yes", "No!") == "Yes":
            cdebug.throwException()

    errmenu.addItem(OOFMenuItem('CError', callback=_cerror,
                                 threadable=oofmenu.THREADABLE))

    def _cpyerror(menuitem):
        debug.fmsg()
        if reporter.query("Really?", "Yes", "No!") == "Yes":
            cdebug.throwPythonException()

    errmenu.addItem(OOFMenuItem('CPyError', callback=_cpyerror,
                                threadable=oofmenu.THREADABLE))

    def _cpycerror(menuitem):
        debug.fmsg()
        if reporter.query("Really?", "Yes", "No!") == "Yes":
            cdebug.throwPythonCException()

    errmenu.addItem(OOFMenuItem("CPyCError", callback=_cpycerror,
                                threadable=oofmenu.THREADABLE))
    

    def loop(menuitem):
        while 1:
            pass
        debug.fmsg("What am I doing here?")

    errmenu.addItem(OOFMenuItem('Infinite_Loop', callback=loop,
                                 threadable=oofmenu.THREADABLE,
                                 help="I hope you have lots of time."))

    import os
    from oof2.SWIG.common import lock
    import time

    rw = lock.RWLock()

    def _py_read(menuitem, seconds):
        global rw
        rw.read_acquire()
        print "Got read permission for %d seconds." % seconds
        time.sleep(seconds)
        print "Releasing read."
        rw.read_release()

    def _py_write(menuitem, seconds):
        global rw
        rw.write_acquire()
        print "Got write permission for %d seconds." % seconds
        time.sleep(seconds)
        print "Releasing write."
        rw.write_release()

    debugmenu.addItem(OOFMenuItem('RWLock_read', callback=_py_read,
                                  no_doc=1,
                                  threadable=oofmenu.THREADABLE,
                                  params=[IntParameter('seconds', 10,
                                                       tip="Sleeping time.")],
                                  help='Safe when used as directed.  For entertainment purposes only.'))

    debugmenu.addItem(OOFMenuItem('RWLock_write', callback=_py_write,
                                  no_doc=1,
                                  threadable=oofmenu.THREADABLE,
                                  params=[IntParameter('seconds', 10,
                                                       tip="Sleeping time.")],
                                  help='Packaged by weight, contents may settle during shipping.'))


    def _wait(menuitem, seconds):
        cdebug.wait(seconds)

    debugmenu.addItem(OOFMenuItem('Wait', callback=_wait,
                                  no_doc=1,
                                  threadable=oofmenu.THREADABLE,
                                  params=[IntParameter('seconds', 10,
                                                       tip="Waiting time.")],
                                  help='For internal use only.'))

