# -*- python -*-
# $RCSfile: imagePage.py,v $
# $Revision: 1.55 $
# $Author: reida $
# $Date: 2006/04/28 19:02:49 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from oof2.SWIG.common import switchboard
from oof2.SWIG.image import oofimage
from oof2.common import debug
from oof2.common import guitop
from oof2.common import labeltree
from oof2.common import microstructure
from oof2.common import ringbuffer
from oof2.common import utils
from oof2.common.IO import mainmenu
from oof2.common.IO import parameter
from oof2.common.IO import whoville
from oof2.common.IO import reporter
from oof2.common.IO.GUI import chooser
from oof2.common.IO.GUI import fileselector
from oof2.common.IO.GUI import gtklogger
from oof2.common.IO.GUI import gtkutils
from oof2.common.IO.GUI import historian
from oof2.common.IO.GUI import oofGUI
from oof2.common.IO.GUI import parameterwidgets
from oof2.common.IO.GUI import regclassfactory
from oof2.common.IO.GUI import whowidget
from oof2.image import imagemodifier
from types import *
import gtk

#font = load_font("-*-fixed-*-*-*-*-*-120-*-*-*-*-iso8859-*")

class ImagePage(oofGUI.MainPage):
    def __init__(self):
        oofGUI.MainPage.__init__(self, name="Image", ordering=50,
                                 tip='Manipulate Images')

        mainbox = gtk.VBox(spacing=2)
        self.gtk.add(mainbox)

        align = gtk.Alignment(xalign=0.5)
        mainbox.pack_start(align, expand=0, fill=0)
        centerbox = gtk.HBox(spacing=3)
        align.add(centerbox)
        label = gtk.Label('Microstructure=')
        label.set_alignment(1.0, 0.5)
        centerbox.pack_start(label, expand=0, fill=0)
        self.imagewidget = whowidget.WhoWidget(oofimage.imageContexts)
        centerbox.pack_start(self.imagewidget.gtk[0], expand=0, fill=0)
        label = gtk.Label('Image=')
        label.set_alignment(1.0, 0.5)
        centerbox.pack_start(label, expand=0, fill=0)
        centerbox.pack_start(self.imagewidget.gtk[1], expand=0, fill=0)

        align = gtk.Alignment(xalign=0.5)
        mainbox.pack_start(align, expand=0, fill=0)
        centerbox = gtk.HBox(homogeneous=1, spacing=3)
        align.add(centerbox)

        self.loadbutton = gtk.Button('Load...')
        gtklogger.setWidgetName(self.loadbutton, 'Load')
        centerbox.pack_start(self.loadbutton, expand=1, fill=1)
        gtklogger.connect(self.loadbutton, 'clicked', self.loadCB)
        guitop.top().tooltip(self.loadbutton,
                             'Load a new image into an existing Microstructure')
        
        self.copybutton = gtkutils.stockButton(gtk.STOCK_COPY, 'Copy...')
        gtklogger.setWidgetName(self.copybutton, 'Copy')
        gtklogger.connect(self.copybutton, 'clicked', self.copyCB)
        centerbox.pack_start(self.copybutton, expand=1, fill=1)
        guitop.top().tooltip(self.copybutton,
                             'Copy the current image.  The copy can be in the same or a different Microstructure.')

        self.renamebutton = gtkutils.stockButton(gtk.STOCK_EDIT, 'Rename...')
        gtklogger.setWidgetName(self.renamebutton, 'Rename')
        gtklogger.connect(self.renamebutton, 'clicked', self.renameCB)
        guitop.top().tooltip(self.renamebutton,
                             'Rename the current image.')
        centerbox.pack_start(self.renamebutton, expand=1, fill=1)

        self.deletebutton = gtkutils.stockButton(gtk.STOCK_DELETE, 'Delete')
        gtklogger.setWidgetName(self.deletebutton, 'Delete')
        gtklogger.connect(self.deletebutton, 'clicked', self.deleteCB)
        guitop.top().tooltip(self.deletebutton,
                             'Delete the current image.')
        centerbox.pack_start(self.deletebutton, expand=1, fill=1)

        self.savebutton = gtkutils.stockButton(gtk.STOCK_SAVE, 'Save...')
        gtklogger.setWidgetName(self.savebutton, 'Save')
        gtklogger.connect(self.savebutton, 'clicked', self.saveCB)
        guitop.top().tooltip(self.savebutton,
                             'Save the current image to a file.')
        centerbox.pack_start(self.savebutton, expand=1, fill=1)

        self.autogroupbutton = gtk.Button('Group')
        gtklogger.setWidgetName(self.autogroupbutton, 'Group')
        gtklogger.connect(self.autogroupbutton, 'clicked', self.autogroupCB)
        centerbox.pack_start(self.autogroupbutton, expand=1, fill=1, padding=2)
        guitop.top().tooltip(self.autogroupbutton, "Create a pixel group in the current image's microstructure for each color pixel in the image.")

        mainpane = gtk.HPaned()
        gtklogger.setWidgetName(mainpane, 'Pane')
        mainbox.pack_start(mainpane, expand=1, fill=1)

        frame = gtk.Frame('Image Information')
        frame.set_shadow_type(gtk.SHADOW_IN)
        mainpane.pack1(frame, resize=True, shrink=False)
        gtklogger.connect_passive(frame, 'size-allocate')
        scroll = gtk.ScrolledWindow()
        gtklogger.logScrollBars(scroll, "StatusScroll")
        scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        frame.add(scroll)

        self.infoarea = gtk.TextView()
        self.infoarea.set_wrap_mode(gtk.WRAP_WORD)
        self.infoarea.set_editable(False)
        self.infoarea.set_cursor_visible(False)
        scroll.add_with_viewport(self.infoarea)

        frame = gtk.Frame('Image Modification')
        frame.set_shadow_type(gtk.SHADOW_IN)
        mainpane.pack2(frame, resize=False, shrink=False)
        vbox = gtk.VBox()
        frame.add(vbox)
##        scroll = gtk.ScrolledWindow()    # scroll window for image mod method
##        scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
##        vbox.pack_start(scroll, expand=1, fill=1)
        self.imageModFactory = regclassfactory.RegisteredClassFactory(
            imagemodifier.ImageModifier.registry,
            title="Method:", name="Method")
##        scroll.add_with_viewport(self.imageModFactory.gtk)
        vbox.pack_start(self.imageModFactory.gtk, expand=1, fill=1)
        self.historian = historian.Historian(self.imageModFactory.set,
                                             self.sensitizeHistory,
                                             setCBkwargs={'interactive':1})
        self.imageModFactory.set_callback(self.historian.stateChangeCB)

        # Prev, OK, and Next buttons
        hbox = gtk.HBox()
        vbox.pack_start(hbox, expand=0, fill=0, padding=2)
        self.prevmethodbutton = gtkutils.prevButton()
        gtklogger.connect(self.prevmethodbutton, 'clicked',
                          self.historian.prevCB)
        hbox.pack_start(self.prevmethodbutton, expand=0, fill=0, padding=2)
        guitop.top().tooltip(self.prevmethodbutton,
                             'Recall the previous image modification operation.')
        self.okbutton = gtk.Button(stock=gtk.STOCK_OK)
        gtklogger.setWidgetName(self.okbutton, 'OK')
        hbox.pack_start(self.okbutton, expand=1, fill=1, padding=5)
        gtklogger.connect(self.okbutton, 'clicked', self.okbuttonCB)
        guitop.top().tooltip(self.okbutton,
                  'Perform the image modification operation defined above.')
        self.nextmethodbutton = gtkutils.nextButton()
        gtklogger.connect(self.nextmethodbutton, 'clicked',
                          self.historian.nextCB)
        hbox.pack_start(self.nextmethodbutton, expand=0, fill=0, padding=2)
        guitop.top().tooltip(self.nextmethodbutton,
                             "Recall the next image modification operation.")

        # Undo and Redo buttons
        hbox = gtk.HBox()
        vbox.pack_start(hbox, expand=0, fill=0, padding=2)
        self.undobutton = gtk.Button(stock=gtk.STOCK_UNDO)
        gtklogger.setWidgetName(self.undobutton, 'Undo')
        gtklogger.connect(self.undobutton, 'clicked', self.undoCB)
        guitop.top().tooltip(self.undobutton,
                             'Undo the latest image modification.')
        hbox.pack_start(self.undobutton, expand=1, fill=0, padding=10)
        self.redobutton = gtk.Button(stock=gtk.STOCK_REDO)
        gtklogger.setWidgetName(self.redobutton, 'Redo')
        gtklogger.connect(self.redobutton, 'clicked', self.redoCB)
        guitop.top().tooltip(self.redobutton,
                             'Redo the latest undone image modification.')
        hbox.pack_start(self.redobutton, expand=1, fill=0, padding=10)

        self.sensitize()
        self.sensitizeHistory()
        self.displayImageInfo()
        
        self.sbcallbacks = [
            switchboard.requestCallbackMain(('new who', 'Image'),
                                            self.newImageCB),
            switchboard.requestCallbackMain(('remove who', 'Image'),
                                            self.rmWhoCB),
            switchboard.requestCallbackMain('modified image',
                                            self.modifiedImageCB),
            switchboard.requestCallbackMain(imagemodifier.ImageModifier,
                                            self.updateImageModifiers),
            switchboard.requestCallbackMain(self.imagewidget,
                                            self.iWidgetChanged),
            switchboard.requestCallbackMain(('validity',
                                             self.imageModFactory),
                                            self.validityChangeCB),
            switchboard.requestCallbackMain(('WhoDoUndo buffer change',
                                             'Image'),
                                            self.whoBufChangeCB)
            ]

    def sensitize(self):
        image = self.getCurrentImage()
        selected = image is not None
        self.copybutton.set_sensitive(selected)
        self.renamebutton.set_sensitive(selected)
        self.deletebutton.set_sensitive(selected)
        self.savebutton.set_sensitive(selected)
        self.okbutton.set_sensitive(selected and self.imageModFactory.isValid())
        self.autogroupbutton.set_sensitive(selected)
        self.loadbutton.set_sensitive(self.getCurrentMS() is not None)
        if selected:
            self.undobutton.set_sensitive(image.undoable())
            self.redobutton.set_sensitive(image.redoable())
        else:
            self.undobutton.set_sensitive(0)
            self.redobutton.set_sensitive(0)

    def sensitizeHistory(self):
        debug.mainthreadTest()
        self.nextmethodbutton.set_sensitive(self.historian.nextSensitive())
        self.prevmethodbutton.set_sensitive(self.historian.prevSensitive())

    def validityChangeCB(self, validity):
        self.sensitize()

    def whoBufChangeCB(self):
        self.sensitize()

    def iWidgetChanged(self, interactive):
        self.sensitize()
        self.displayImageInfo()
        
    def displayImageInfo(self):
        debug.mainthreadTest()
        imagecontext = self.getCurrentImage()
        text = ""
        if imagecontext:
            image = imagecontext.getObject()
            size = image.sizeInPixels()
            text += 'Pixel size: %dx%d\n' % (size.x, size.y)
            size = image.size()
            text += 'Physical size: %gx%g\n' % (size.x, size.y)
            if image.comment():
                text += '\nComments:\n%s\n' % image.comment()
        self.infoarea.get_buffer().set_text(text)

    def getCurrentImageName(self):
        path = labeltree.makePath(self.imagewidget.get_value())
        if path:
            return path[-1]

    def getCurrentMSName(self):
        path = labeltree.makePath(self.imagewidget.get_value())
        if path:
            return path[0]

    def getCurrentMS(self):
        try:
            return microstructure.microStructures[self.getCurrentMSName()]
        except KeyError:
            pass
        
    def getCurrentImage(self):
        try:
            return oofimage.imageContexts[self.imagewidget.get_value()]
        except KeyError:
            pass
            

    def updateImageModifiers(self):
        self.imageModFactory.update(imagemodifier.ImageModifier.registry)

    def newImageCB(self, whoname):        # switchboard 'new who', 'Image'
        self.imagewidget.set_value(whoname)
        self.sensitize()

    def rmWhoCB(self, whoname):         # switchboard 'remove who'
        self.sensitize()

    ############### GUI callbacks ################

    def loadCB(self, button):
        menuitem = mainmenu.OOF.File.Load.Image
        menuitem.gui_callback(menuitem)

    def copyCB(self, button):
        menuitem = mainmenu.OOF.Image.Copy
        nameparam = menuitem.get_arg('name')
        msparam = menuitem.get_arg('microstructure')
        if parameterwidgets.getParameters(msparam, nameparam,
                                          title="Copy Image"):
            menuitem.callWithDefaults(image=self.imagewidget.get_value())

    def renameCB(self, button):
        menuitem = mainmenu.OOF.Image.Rename
        namearg = menuitem.get_arg('name')
        namearg.value = labeltree.makePath(self.imagewidget.get_value())[-1]
        if parameterwidgets.getParameters(namearg, title='Rename Image'):
            menuitem.callWithDefaults(image=self.imagewidget.get_value())

    def deleteCB(self, button):
        if reporter.query("Really delete %s?"
                          % self.getCurrentImageName(),
                          "No", default="Yes") == "Yes":
            mainmenu.OOF.Image.Delete(image=self.imagewidget.get_value())
            
    def saveCB(self,button):
        file = fileselector.fileSelector(ident='SaveImage', title='Save Image',
                                         mode='w',
                                         filename=self.getCurrentImageName())
        if file is not None:
            mainmenu.OOF.File.Save.Image(image=self.imagewidget.get_value(),
                                         filename=file)

    def okbuttonCB(self, button):
        modmeth = self.imageModFactory.getRegistration()
        if modmeth is not None:
            # buildImageModMenu() in imagemenu.py builds the
            # Image.Modify menu items from the members of the
            # ImageModifier RegisteredClass.  The Parameters in the
            # menu items are identically the Parameters in the
            # Registrations.  So copying new values out of the Factory
            # into the Registrations, and invoking the menu item with
            # callWithDefaults() executes the command with the current
            # arguments.
            self.imageModFactory.set_defaults() # copy from gui to Registration
            menuitem = getattr(mainmenu.OOF.Image.Modify, modmeth.name())
            menuitem.callWithDefaults(image=self.imagewidget.get_value())

    def modifiedImageCB(self, imageModifier, imagename): # sb 'modified image'
        # Called whenever the image is changed, either by applying an
        # ImageModifier or the Undo and Redo buttons, in which case
        # the imageModifier arg is None.
        if imageModifier is not None:
            self.historian.record(imageModifier)
        self.sensitize()
        
    def undoCB(self, button):
        mainmenu.OOF.Image.Undo(image=self.imagewidget.get_value())

    def redoCB(self, button):
        mainmenu.OOF.Image.Redo(image=self.imagewidget.get_value())

    def autogroupCB(self, button):
        mainmenu.OOF.Image.AutoGroup(image=self.imagewidget.get_value())
        
ImagePage()

###########################

def _saveImage(menuitem):
    file = fileselector.fileSelector(
        mode="w",
        ident="SaveImage",
        title="Save Image",
        params=[menuitem.get_arg('image')])
    if file is not None:
        menuitem.callWithDefaults(filename=file)

mainmenu.OOF.File.Save.Image.add_gui_callback(_saveImage)
